// EV3DirectServer.java

/* History:
  V1.00-V1.30: Beta releases
  V1.31 - Jul 2014: - New EV3JLibA library (version 1.02)
  V1.32 - Jul 2014: - New EV3JLibA library (version 1.03)
  V1.33 - Jul 2014: - New EV3JLibA library (version 1.04)
*/

import ch.aplu.ev3.*;
import java.io.*;
import java.lang.reflect.Method;
import java.net.*;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import lejos.hardware.LED;
import lejos.hardware.Brick;
import lejos.hardware.Sound;
import lejos.hardware.ev3.LocalEV3;
import lejos.hardware.lcd.LCD;
import org.apache.commons.codec.binary.Base64;

public class EV3DirectServer extends Thread
{
  // -------------- Inner class Flasher ----------------------
  private class Flasher extends Thread
  {
    public void run()
    {
      while (isFlashing)
      {
        led.setPattern(1);
        delay(50);
        led.setPattern(0);
        delay(4000);
      }
      led.setPattern(0);
    }

    void delay(long time)
    {
      try
      {
        Thread.sleep(time);
      }
      catch (InterruptedException ex)
      {
      }
    }
  }

  // ------------- Inner class DynamicUpdateClient ----
  private class DynamicUpdateClient extends Thread
  {
    private String hostUrl;
    private String ev3Url;
    private String userpass;
    private int ducInterval;
    private int ducIdleInterval;
    private boolean isFirst = true;

    public DynamicUpdateClient(String hostUrl, String ev3Url,
      String userpass, int ducInterval, int ducIdleInterval)
    {
      this.hostUrl = hostUrl;
      this.ev3Url = ev3Url;
      this.userpass = userpass;
      this.ducInterval = ducInterval;
      this.ducIdleInterval = ducIdleInterval;
    }

    public void run()
    {
      LCD.drawString("DUC started", 0, 7);
      while (isDucActive)
      {
        if (isFirst)
        {
          isFirst = false;
          updateRouterAddress(true);
        }
        else
          updateRouterAddress(!isConnected);  // Don't show if connected
        if (isDucActive)
        {
          try
          {
            if (isConnected)
              Thread.currentThread().sleep(60000 * ducInterval);
            else
              Thread.currentThread().sleep(60000 * ducIdleInterval);
          }
          catch (InterruptedException ex)
          {
          }
        }
      }
      showStatus("DUC terminated", "");
    }

    private void updateRouterAddress(boolean showStatus)
    {
      String params = "?hostname=" + ev3Url;
      try
      {
        URL dynupdate
          = new URL(hostUrl + params);
        URLConnection uc = dynupdate.openConnection();
        String basicAuth = "Basic "
          + new String(new Base64().encode(userpass.getBytes()));
        uc.setRequestProperty("Authorization", basicAuth);
        BufferedReader in = new BufferedReader(
          new InputStreamReader(
            uc.getInputStream()));
        String msg = "";
        String inputLine;
        int i = 0;
        while ((inputLine = in.readLine()) != null)
        {
          if (i == 0)
            msg = inputLine;
          i++;
        }
        in.close();
        if (showStatus)
          showStatus("DUC reply:", msg);
      }
      catch (IOException ex)
      {
        isDucActive = false;
        showStatus("DUC disabled", null);
      }
    }
  }

  // ------------- Inner interface Response -----------
  private interface Response
  {
    String OK = "0";
    String SEND_FAILED = "-1";
    String ILLEGAL_METHOD = "-2";
    String ILLEGAL_INSTANCE = "-3";
    String CMD_ERROR = "-4";
    String ILLEGAL_PORT = "-5";
    String CREATION_FAILED = "-6";
  }

  // ------------- End of inner classes ------------------
  //
  private final HashMap<String, String> classTags = new HashMap<String, String>()
  {
    // Relation between device name tag transmitted from client and class name

    
    {
      put("cs", "ColorSensor");
      put("_cs", "NxtColorSensor");
      put("gear", "Gear");
      put("_gear", "NxtGear");
      put("grs", "GyroRateSensor");
      put("gas", "GyroAngleSensor");
      put("htas", "HTAccelerometer");
      put("htgs", "HTGyroSensor");
      put("htbs", "HTBarometer");
      put("htcp", "HTCompassSensor");
      put("htcs", "HTColorSensor");
      put("htis", "HTInfraredSeeker");
      put("htis$", "HTInfraredSeekerV2");
      put("ird", "IRDistanceSensor");
      put("irr", "IRRemoteSensor");
      put("irs", "IRSeekSensor");
      put("ls", "LightSensor");
      put("_ls", "NxtLightSensor");
      put("mot", "Motor");
      put("mmot", "MediumMotor");
      put("_mot", "NxtMotor");
      put("ods", "OpticalDistanceSensor");
      put("pts", "PrototypeSensor");
      put("rfid", "RFIDSensor");
      put("sps", "SuperProSensor");
      put("_ss", "NxtSoundSensor");
      put("ts", "TouchSensor");
      put("_ts", "NxtTouchSensor");
      put("us", "UltrasonicSensor");
      put("_us", "NxtUltrasonicSensor");
    }
  };

  private final static String VERSION = "1.33-July 2014";
  private boolean isConnected = false;
  private boolean isExiting = false;
  private int port;
  private ServerSocket serverSocket;
  private Socket socket;
  private InputStream is;
  private OutputStream os;
  private LegoRobot robot;
  private HashMap<String, Object> deviceTable = new HashMap<String, Object>();
  private boolean inError = false;
  private boolean isDucActive = true;
  private DynamicUpdateClient duc = null;
  private LED led;
  private boolean isFlashing = false;
  private Flasher flasher = null;

  public EV3DirectServer()
  {
    LCD.drawString("EV3DirectServer", 0, 0);
    LCD.drawString("(V" + VERSION + ")", 0, 1);
    LCD.refresh();
    Brick localBrick = LocalEV3.get();
    led = localBrick.getLED();

    EV3Properties props = new EV3Properties();
    port = props.getIntValue("IPport");
    isDucActive = props.getStringValue("DucEnabled").equals("true");
    if (isDucActive)
    {
      String hostUrl = props.getStringValue("DucHostUrl");
      String ev3Url = props.getStringValue("DucEV3Url");
      String userpass = props.getStringValue("DucUserPass");
      int ducInterval = props.getIntValue("DucUpdateInterval");
      int ducIdleInterval = props.getIntValue("DucIdleUpdateInterval");
      duc = new DynamicUpdateClient(hostUrl, ev3Url, userpass,
        ducInterval, ducIdleInterval);
      duc.start();
    }
    playServerUpMelody();
    startFlasher();
    start();
    while (!Tools.isEscapePressed())
      Tools.delay(100);
    stopDuc();
    isExiting = true;
    LCD.clear();
    LCD.drawString("Closing now...", 0, 1);
    LCD.refresh();
    stopFlasher();
    led.setPattern(2);
    Tools.delay(2000);
    if (isConnected)
    {
      try
      {
        is.close();
      }
      catch (Exception ex)
      {
      }
    }
    else
    {
      try
      {
        serverSocket.close();
      }
      catch (Exception ex)
      {
      }
    }
    led.setPattern(0);
  }

  public void run()
  {
    String cmd;
    try
    {
      serverSocket = new ServerSocket(port);
      List<String> ips = getIPAddresses();
      if (ips == null)
        throw new IOException();

      LCD.drawString("Listening on IP", 0, 2);
      if (ips.size() == 1)
      {
        LCD.drawString(ips.get(0), 0, 3);
        LCD.drawString("port " + port + "...", 0, 4);
      }
      else if (ips.size() == 2)
      {
        LCD.drawString(ips.get(0), 0, 3);
        LCD.drawString(ips.get(1), 0, 4);
        LCD.drawString("port " + port + "...", 0, 5);
      }
      else if (ips.size() == 3)
      {
        LCD.drawString(ips.get(0), 0, 3);
        LCD.drawString(ips.get(1), 0, 4);
        LCD.drawString(ips.get(2), 0, 5);
        LCD.drawString("port " + port + "...", 0, 6);
      }
      LCD.refresh();
      while (!isExiting)
      {
        isConnected = false;
        socket = serverSocket.accept(); // Blocking
        init();
        stopFlasher();
        led.setPattern(1);
        isConnected = true;
        String ipAddress = socket.getInetAddress().getHostAddress();
        LCD.clear();
        LCD.drawString("Client with IP", 0, 1);
        LCD.drawString(ipAddress, 0, 2);
        LCD.drawString("connected", 0, 3);
        LCD.refresh();
        try
        {
          is = new DataInputStream(socket.getInputStream());
          os = new DataOutputStream(socket.getOutputStream());
          while (true)
          {
            cmd = readCommand();
            executeCommand(cmd);
          }
        }
        catch (Exception ex)
        {
          if (!isExiting)
          {
            LCD.clear();
            LCD.drawString("Disconnected.", 0, 1);
            LCD.drawString("Wait for client...", 0, 2);
            LCD.refresh();
            startFlasher();
            if (isDucActive)
              duc.interrupt();  // Take out of sleep
          }
          if (robot != null)
          {
            robot.exit();
            robot = null;
          }
          cleanup(isExiting);
        }
      }
    }
    catch (IOException ex)
    {
      if (!isExiting)
      {
        stopDuc();
        LCD.clear();
        LCD.drawString("Can't create server", 0, 1);
        LCD.drawString("at port:", 0, 2);
        LCD.drawString("" + port, 0, 3);
        LCD.refresh();
      }
    }
  }

  private void init()
  {
    deviceTable.clear();
    robot = null;
    is = null;
    os = null;
  }

  private void stopDuc()
  {
    if (isDucActive)
    {
      isDucActive = false;
      if (duc != null)
        duc.interrupt();
    }
  }

  private void cleanup(boolean mustCloseServer)
  {
    try
    {
      os.close();
    }
    catch (Exception ex)
    {
    }
    try
    {
      is.close();
    }
    catch (Exception ex)
    {
    }
    try
    {
      socket.close();
    }
    catch (Exception ex)
    {
    }
    if (mustCloseServer)
    {
      stopDuc();
      try
      {
        serverSocket.close();
      }
      catch (Exception ex)
      {
      }
    }
  }

  private void executeCommand(String cmd)
  {
    if (cmd.equals("getVersion"))
    {
      sendReply(VERSION);
      return;
    }

    if (cmd.equals("getClassIdentifiers"))
    {
      String entries = "";
      for (String key : classTags.keySet())
        entries += key + "->" + classTags.get(key) + ";";
      sendReply(entries);
      return;
    }

    if (cmd.equals("getInstanceNames"))
    {
      String entries = "";
      for (String key : deviceTable.keySet())
        entries += key + "->" + classTags.get(key.substring(0, key.length() - 1))
          + ";";
      if (entries.equals(""))
        entries = "empty";
      sendReply(entries);
      return;
    }

    String[] splitter = cmd.split("\\.");  // Split on period .
    if (splitter == null || splitter.length < 2 || splitter.length > 4)
    {
      showError("COMMAND_ERROR", cmd);
      sendReply(Response.CMD_ERROR);   // Command error
      return;
    }
    String[] parts = new String[4];
    parts[0] = splitter[0];
    parts[1] = splitter[1];
    if (splitter.length == 2)
    {
      parts[2] = "n";
      parts[3] = "n";
    }
    else if (splitter.length == 3)
    {
      parts[2] = splitter[2];
      parts[3] = "n";
    }
    else
    {
      parts[2] = splitter[2];
      parts[3] = splitter[3];
    }

    String device = parts[0];
    String method = parts[1];
    String param1 = parts[2];
    String param2 = parts[3];

    if (device.substring(0, device.length() - 1).equals("robo"))
      dispatchRobot(method, param1, param2);
    else
      dispatchCommand(device, method, param1, param2);
  }

  private void dispatchRobot(String method, String param1, String param2)
  {
    String reply = "";
    if (method.equals("create"))
    {
      if (param1.equals("0"))
        robot = new LegoRobot(false, false);
      else
        robot = new LegoRobot(false, true);
      reply = Response.OK;
    }
    else if (method.equals("addPart")) // Special method "addPart"
    {
      String deviceID = param1;
      String port = param2;
      int portID = 0;
      boolean isMotor = false;
      boolean isGear = false;
      try
      {
        if (deviceID.equals("gear") || deviceID.equals("_gear"))
          isGear = true;
        else
        {
          char portChar = port.charAt(0);
          if (portChar > 64 && portChar < 69)
          {
            portID = portChar - 65;
            isMotor = true;
          }
          else
            portID = portChar - 49;

          if (portID < 0 || portID > 3)
          {
            showError("ILLEGAL_PORT", port);
            sendReply(Response.ILLEGAL_PORT);
            return;
          }
        }

        String classNameStr = "ch.aplu.ev3." + classTags.get(deviceID);
        Class deviceClass = Class.forName(classNameStr);
        Object deviceObj;
        if (isMotor)
        {
          deviceObj = deviceClass.getConstructor(MotorPort.class).
            newInstance(new MotorPort(portID));
          deviceTable.put(deviceID + port, deviceObj);
        }
        else if (isGear)
        {
          deviceObj = deviceClass.newInstance();
          deviceTable.put(deviceID, deviceObj);
        }
        else
        {
          deviceObj = deviceClass.getConstructor(SensorPort.class).
            newInstance(new SensorPort(portID));
          deviceTable.put(deviceID + port, deviceObj);
        }
        robot.addPart((Part)deviceObj);
        reply = Response.OK;
      }
      catch (Exception ex)
      {
        showError("CREATION_FAILED", deviceID);
        sendReply(Response.CREATION_FAILED);
        return;
      }
    }
    else  // Standard robot command
    {
      try
      {
        Class c = Class.forName("ch.aplu.ev3.LegoRobot");
        Method m = c.getMethod(method, getArgTypes(param1, param2));
        Object rc = m.invoke(robot, getArgValues(param1, param2));
        if (rc instanceof Integer)
          reply = ((Integer)rc).toString();
        else if (rc instanceof String)
          reply = (String)rc;
        else if (rc instanceof Boolean)
        {
          Boolean b = (Boolean)rc;
          reply = (b ? "1" : "0");
        }
        else
          reply = Response.OK;
      }
      catch (Exception ex)
      {
        showError("ILLEGAL_METHOD", method);
        sendReply(Response.ILLEGAL_METHOD);
        return;
      }
    }
    sendReply(reply);
    clearError();
  }

  private void dispatchCommand(String deviceName, String method, String param1, String param2)
  {
    boolean isMotor = false;
    boolean isGear = false;
    int portId = 0;
    if ((deviceName.length() >= 3 && deviceName.substring(0, 3).equals("mot"))
      || ((deviceName.length() >= 4) && deviceName.substring(1, 4).equals("mot")))
      isMotor = true;

    if (deviceName.equals("gear") || deviceName.equals("_gear"))
      isGear = true;
    else
    {
      if (isMotor)
        portId = (int)deviceName.charAt(deviceName.length() - 1) - 65;
      else
        portId = (int)deviceName.charAt(deviceName.length() - 1) - 49;

      if (portId < 0 || portId > 3)
      {
        showError("ILLEGAL_PORT", deviceName);
        sendReply(Response.ILLEGAL_PORT);
        return;
      }
    }

    String classTag;
    if (isGear)
      classTag = deviceName;
    else
      classTag = deviceName.substring(0, deviceName.length() - 1);

    // For performance optimization, 'getValue' is treated specially
    if (method.equals("getValue"))
    {
      try
      {
        Object deviceObj = deviceTable.get(deviceName);
        Class c = Class.forName("ch.aplu.ev3." + classTags.get(classTag));
        Method m = c.getMethod(method, new Class[]
        {
        });
        Object rc = m.invoke(deviceObj, new Object[]
        {
        });
        sendResponse(rc);
        return;
      }
      catch (Exception ex)
      {
        showError("ILLEGAL_METHOD", method);
        sendReply(Response.ILLEGAL_METHOD);
        return;
      }
    }
    else if (method.equals("readPrototype"))  // Prototype sensor read inputs
    {
      PrototypeSensor ps = (PrototypeSensor)deviceTable.get(deviceName);
      int[] ain = new int[5];
      int[] din = new int[6];
      ps.read(ain, din);
      String reply = "";
      for (int i = 0; i < 5; i++)
        reply += ((Integer)ain[i]).toString() + ":";
      int power = 1;
      int digit = 0;
      for (int i = 0; i < 6; i++)
      {
        digit += din[i] * power;
        power *= 2;
      }
      reply += ((Integer)digit).toString();
      clearError();
      sendReply(reply);
    }
    else if (method.equals("readSuperPro"))  // SuperPrototype sensor read inputs
    {
      SuperProSensor sps = (SuperProSensor)deviceTable.get(deviceName);
      int[] ain = new int[4];
      int[] din = new int[8];
      sps.read(ain, din);
      String reply = "";
      for (int i = 0; i < 4; i++)
        reply += ((Integer)ain[i]).toString() + ":";
      int power = 1;
      int digit = 0;
      for (int i = 0; i < 8; i++)
      {
        digit += din[i] * power;
        power *= 2;
      }
      reply += ((Integer)digit).toString();
      clearError();
      sendReply(reply);
    }
    else // Standard method call
    {
      try
      {
        Object deviceObj = deviceTable.get(deviceName);
        Class c = Class.forName("ch.aplu.ev3." + classTags.get(classTag));
        Method m = c.getMethod(method, getArgTypes(param1, param2));
        Object rc = m.invoke(deviceObj, getArgValues(param1, param2));
        sendResponse(rc);
      }
      catch (Exception ex)
      {
        showError("ILLEGAL_METHOD", method);
        sendReply(Response.ILLEGAL_METHOD);
      }
    }
  }

  private Class[] getArgTypes(String param1, String param2)
  {
    Class[] argTypes;
    if (param1.equals("n"))
    {
      argTypes = new Class[]
      {
      };
    }
    else if (param2.equals("n"))
    {
      if (param1.charAt(0) == 'b')
      {
        argTypes = new Class[]
        {
          boolean.class
        };
      }
      else
      {
        argTypes = new Class[]
        {
          int.class
        };
      }
    }
    else
    {
      if (param2.charAt(0) == 'b')
      {
        argTypes = new Class[]
        {
          int.class, boolean.class
        };
      }
      else
      {
        argTypes = new Class[]
        {
          int.class, int.class
        };
      }
    }
    return argTypes;
  }

  private Object[] getArgValues(String param1, String param2)
  {
    Object[] argValues = null;
    if (param1.equals("n"))
    {
      argValues = new Object[]
      {
      };
    }
    else if (param2.equals("n"))
    {
      if (param1.charAt(0) == 'b')
      {
        argValues = new Object[]
        {
          (param1.charAt(1) == '1')
        };
      }
      else
      {
        argValues = new Object[]
        {
          new Integer(param1)
        };
      }
    }
    else
    {
      if (param2.charAt(0) == 'b')
      {
        argValues = new Object[]
        {
          new Integer(param1), (param2.charAt(1) == '1')
        };
      }
      else
      {
        argValues = new Object[]
        {
          new Integer(param1), new Integer(param2)
        };
      }
    }
    return argValues;
  }

  private void sendResponse(Object rc)
  {
    String reply;
    if (rc instanceof Integer)
      reply = ((Integer)rc).toString();
    else if (rc instanceof Long)
      reply = ((Long)rc).toString();
    else if (rc instanceof String)
      reply = (String)rc;
    else if (rc instanceof Boolean)
    {
      Boolean b = (Boolean)rc;
      reply = (b ? "1" : "0");
    }
    else if (rc instanceof Vector2D)
    {
      Vector2D v = (Vector2D)rc;
      reply = v.x + ";" + v.y;
    }
    else if (rc instanceof Vector3D)
    {
      Vector3D v = (Vector3D)rc;
      reply = v.x + ";" + v.y + ";" + v.z;
    }
    else
      reply = Response.OK;
    clearError();
    sendReply(reply);
  }

  private String readCommand() throws IOException
  {
    byte[] buf = new byte[4096];
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    boolean done = false;
    while (!done)
    {
      int len = is.read(buf);
      if (len == -1)
        throw new IOException("Stream closed");
      baos.write(buf, 0, len);
      if (buf[len - 1] == 10)  // \n
        done = true;
    }
    String s = baos.toString("UTF-8");
    return s.substring(0, s.length() - 1);  // Remove \n
  }

  private void sendReply(String reply)
  {
    reply = reply + "\n";
    byte[] ary = reply.getBytes(Charset.forName("UTF-8"));
    try
    {
      os.write(ary);
      os.flush();
    }
    catch (IOException ex)
    {
      showError("sendReply()", "failed");
    }
  }

  private void showError(String msg1, String msg2)
  {
    inError = true;
    LCD.clear(5);
    LCD.clear(6);
    LCD.drawString(msg1, 0, 5);
    LCD.drawString(msg2, 0, 6);
    LCD.refresh();
  }

  private void showStatus(final String msg1, final String msg2)
  {
    new Thread()
    {
      public synchronized void run()
      {
        LCD.clear(6);
        LCD.drawString(msg1, 0, 6);
        if (msg2 != null)
        {
          LCD.clear(7);
          LCD.drawString(msg2, 0, 7);
        }
        LCD.refresh();
        Tools.delay(4000);
        LCD.clear(6);
        if (msg2 != null)
          LCD.clear(7);
        LCD.refresh();
      }
    }.start();
  }

  private void clearError()
  {
    // Clear only, if an error was displayed, because of delay
    // for displaying text  
    if (inError)
    {
      inError = false;
      LCD.clear(5);
      LCD.clear(6);
      LCD.refresh();
    }
  }

  private static void showDebug(String msg)
  {
    LCD.clear();
    LCD.refresh();
    System.out.println(msg);
    Tools.delay(6000);
  }

  private void playServerUpMelody()
  {
    Sound.playTone(2000, 50);
    Tools.delay(30);
    Sound.playTone(2000, 50);
    Tools.delay(30);
  }

  private void playDebugMelody()
  {
    Sound.playTone(500, 100);
    Tools.delay(100);
  }

  /**
   * Get all the IP addresses for the device
   */
  public static List<String> getIPAddresses()
  {
    List<String> result = new ArrayList<String>();
    Enumeration<NetworkInterface> interfaces;
    try
    {
      interfaces = NetworkInterface.getNetworkInterfaces();
      while (interfaces.hasMoreElements())
      {
        NetworkInterface current = interfaces.nextElement();
        if (!current.isUp() || current.isLoopback() || current.isVirtual())
          continue;
        Enumeration<InetAddress> addresses = current.getInetAddresses();
        while (addresses.hasMoreElements())
        {
          InetAddress current_addr = addresses.nextElement();
          if (current_addr.isLoopbackAddress())
            continue;
          result.add(current_addr.getHostAddress());
        }
      }
      return result;
    }
    catch (Exception ex)
    {
      System.out.println(ex);
      return null;
    }
  }

  private void startFlasher()
  {
    if (flasher != null)
      return;
    isFlashing = true;
    flasher = new Flasher();
    flasher.start();
  }

  private void stopFlasher()
  {
    if (flasher == null)
      return;
    isFlashing = false;
    flasher.interrupt();
    while (flasher.isAlive())
      Tools.delay(1);
    flasher = null;
  }

  public static void main(String[] args)
  {
    new EV3DirectServer();
  }
}
