# LegoRobot.py
# Even in direct mode from remote, the key events are generated by EV3 buttons

import socket, sys, time
from threading import Thread, Lock
from BrickButton import BrickButton

isTkinterAvailable = True
try:
   import Tkinter as tkinter
   import tkMessageBox as mbox
except ImportError:
   isTkinterAvailable = False

# -------------------------------- global constants -------------------------
receiverTimeout = 10
VERSION = "V1.33"
DEBUG = False  # Be careful: Too many debugging messages may influence running behavior when program is spawned from remote jSch SSH
USE_DIALOG = True  # Connection box

# -------------------------------- class Response ---------------------------
class Response():
    OK = "0"
    SEND_FAILED = "-1"
    ILLEGAL_METHOD = "-2"
    ILLEGAL_INSTANCE = "-3"
    CMD_ERROR = "-4"
    ILLEGAL_PORT = "-5"
    CREATION_FAILED = "-6"

# -------------------------------- class ModelessMessageBox -----------------
class ModelessMessageBox(Thread):
    def __init__(self, robot, msg):
        Thread.__init__(self)
        self.robot = robot
        self.msg = msg

    def run(self):
        self.window = tkinter.Tk()
        self.window.wm_withdraw()
        mbox.showinfo("EV3PyLib " + VERSION, self.msg)
        self.robot.exit()

# -------------------------------- class MyKeyListener ---------------------
class MyKeyListener(Thread):
    pollPeriod = 0.2 # s    
    
    def __init__(self, robot, buttonHit):
        Thread.__init__(self)
        self.robot = robot
        self.buttonHit = buttonHit
                  
    def run(self):
        while self.robot.isKeyListenerRunning:
           keyState = int(self.robot.sendCommand("robot.getHitButtonID"))
           if keyState > 0:
               self.robot.buttonID = keyState
               if self.buttonHit != None:
                   self.buttonHit(keyState)
           time.sleep(self.pollPeriod)
       
# -------------------------------- class Receiver ---------------------------
class Receiver(Thread):
    def __init__(self, robot):
        Thread.__init__(self)
        self.robot = robot
                  
    def run(self):
        LegoRobot.showDebug("Receiver thread started")
        while self.robot.isReceiverRunning:
            try:
                self.robot.isReceiverUp = True
                self.robot.receiverResponse = self.readResponse()
            except:    
                LegoRobot.showDebug("LegoRobot exception in Receiver.run(): " + str(sys.exc_info()[1]))
                self.robot.isReceiverRunning = False
                self.robot.closeConnection()
        LegoRobot.showDebug("Receiver thread terminated")

    def readResponse(self):
        bufSize = 4096
        data = ""
        while not "\n" in data:
            try:
               reply = self.robot.sock.recv(bufSize)
            except:
                raise Exception("Exception from blocking sock.recv()")
                
            data += reply
        data = data.decode("UTF-8")    
        data = data[:-1]  # Remove trailing \n
        return data

# -------------------------------- class LegoRobot --------------------------
class LegoRobot():
    def __init__(self, buttonHit = None, *args):
        global debug
        if len(args) == 0:
            self.ipAddress = "10.0.1.1"
            self.port = 1299
        elif len(args) == 1:
            self.ipAddress = args[0]
            self.port = 1299
        elif len(args) == 2:
            self.ipAddress = args[0]
            self.port = args[1]
            debug = False
        elif len(args) == 3:
            self.ipAddress = args[0]
            self.port = args[1]
        else:
            raise ValueError("Illegal number of parameters")
        
        self.isReceiverUp = False
        self.isReceiverRunning = False
        self.isConnected = False
        self.buttonID = 0
        self.isKeyListenerRunning = False
        self.buttonHit = buttonHit
        self.parts = []
        self.lock = Lock()
        if self.connect():
            self.sendCommand("robot.create.1")
            self.initKeyListener()
            if USE_DIALOG:
                if isTkinterAvailable:
                    self.mmb = ModelessMessageBox(self, "Press OK to quit")
                    self.mmb.start()
        else:
            raise Exception("Connection failed.")

    @staticmethod
    def showDebug(msg):
        if DEBUG:
           print msg        

    def startReceiver(self):
        LegoRobot.showDebug("Starting Receiver thread")
        self.isReceiverRunning = True
        receiver = Receiver(self)
        receiver.start()
        while not self.isReceiverUp:
           time.sleep(0.001)
        time.sleep(0.1)

    def sendCommand(self, cmd):
        self.lock.acquire()
        LegoRobot.showDebug("sendCommand() with cmd = " + cmd)
        if not self.isConnected:
            LegoRobot.showDebug("Not connected. Returned: SEND_FAILED")
            return Response.SEND_FAILED

        try:
            self.receiverResponse = None
            cmd += "\n";  # Append \n
            self.sock.sendall(cmd)
            reply = self.waitForReply()  # Throws exception if timeout
            self.lock.release()
            return reply    
        except:
            LegoRobot.showDebug("LegoRobot exception in sendCommand(): " + str(sys.exc_info()[1]))
            self.closeConnection()

        LegoRobot.showDebug("Not connected. Returned: SEND_FAILED")
        self.lock.release()
        return Response.SEND_FAILED

    def waitForReply(self):
        LegoRobot.showDebug("Calling waitForReply")
        startTime = time.clock()
        while self.isConnected and self.receiverResponse == None and time.clock() - startTime < receiverTimeout:
            time.sleep(0.01)
        if self.receiverResponse == None:
            raise Exception("Exception in LegoRobot: waitForReply failed.")
        LegoRobot.showDebug("Response = " + self.receiverResponse)
        return self.receiverResponse
 
    def closeConnection(self):
        if not self.isConnected:
            return
        self.isConnected = False
        LegoRobot.showDebug("Closing socket")
        self.sock.shutdown(socket.SHUT_WR)
        self.sock.close()
    
    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        LegoRobot.showDebug("Connecting...")
        try:
            self.sock.connect((self.ipAddress, self.port))
            self.isConnected = True
        except:
            LegoRobot.showDebug("Connection failed.")
            return False

        LegoRobot.showDebug("Connection OK")
        self.startReceiver()
        return True

    def reset(self):
        pass
  
    def exit(self):
        self.stopKeyListener()
        for part in self.parts:
            part.cleanup()  
        self.closeConnection()
  
    def isConnected(self):
        return self.isConnected
    
    def playTone(self, frequency, duration):
        self.sendCommand("robot.playTone." + str(int(frequency)) + "." + str(int(duration)))

    def setVolume(self, volume):
        self.sendCommand("robot.setVolume." + str(int(volume)))

    def setLED(self, pattern):
        self.sendCommand("robot.setLED." + str(int(pattern)))
   
    def getBatteryLevel(self):
        return int(self.sendCommand("robot.getBatteryLevel"))

    def addPart(self, part):
        part.setRobot(self)
        portName = part.partName[len(part.partName) - 1]
        if portName == 'A' or portName == 'B' or portName == 'C' or portName == 'D' or portName == '1' or portName == '2' or portName == '3' or portName == '4':
            classID = part.partName[:-1]
            self.sendCommand("robot.addPart." + classID + "." + portName)
        else:
            self.sendCommand("robot.addPart." + part.partName)
        part.init()
        self.parts.append(part)
        
    def initKeyListener(self):
        if self.isKeyListenerRunning:
           return
        self.buttonID = 0
        self.isKeyListenerRunning = True
        self.keyListener = MyKeyListener(self, self.buttonHit)
        self.keyListener.start()

    def stopKeyListener(self):
        if not self.isKeyListenerRunning:
            return
        self.buttonID = 0
        self.isKeyListenerRunning = False

    def isButtonHit(self):
       time.sleep(0.01)
       if not self.isConnected:
          return False
       return (self.buttonID != 0)

    def getHitButtonID(self):
       if not self.isConnected:
           return 0
       id = self.buttonID
       self.buttonID = 0
       return id

    def isUpHit(self):
        time.sleep(0.01)
        if not self.isConnected:
            return False
        pressed = (self.buttonID == BrickButton.ID_UP)
        if pressed:
            self.buttonID = 0
        return pressed

    def isDownHit(self):
        time.sleep(0.01)
        if not self.isConnected:
            return False
        pressed = (self.buttonID == BrickButton.ID_DOWN)
        if pressed:
            self.buttonID = 0
        return pressed

    def isLeftHit(self):
        time.sleep(0.01)
        if not self.isConnected:
            return False
        pressed = (self.buttonID == BrickButton.ID_LEFT)
        if pressed:
            self.buttonID = 0
        return pressed

    def isRightHit(self):
        time.sleep(0.01)
        if not self.isConnected:
            return False
        pressed = (self.buttonID == BrickButton.ID_RIGHT)
        if pressed:
            self.buttonID = 0
        return pressed

    def isEnterHit(self):
        time.sleep(0.01)
        if not self.isConnected:
            return False
        pressed = (self.buttonID == BrickButton.ID_ENTER)
        if pressed:
            self.buttonID = 0
        return pressed

    def isEscapeHit(self):
        time.sleep(0.01)
        if not self.isConnected:
            return False
        pressed = (self.buttonID == BrickButton.ID_ESCAPE)
        if pressed:
            self.buttonID = 0
        return pressed

    def drawString(self, text, x, y):
        self.drawStringAt(text, 18 * y + x)

    def drawStringAt(self, text, count):
        text = text.replace(".", "`")
        self.sendCommand("robot.drawStringAt.s" + text + "." + str(count))
  
    def clearDisplay(self):
        self.sendCommand("robot.clearDisplay")
  
    def getBrickGateVersion(self):
        return self.sendCommand("getBrickGateVersion")

    def getIPAddresses(self):
        response = self.sendCommand("getIPAddresses")
        return response.split(";")

    def isAutonomous(self):
        response = self.sendCommand("isAutonomous")
        if response == "1":
            return True
        return False 
       
