// TargetCopy.java
// Direct mode

/*
 This software is part of the EV3JLib library.
 It is Open Source Free Software, so you may
 - run the code for any purpose
 - study how the code works and adapt it to your needs
 - integrate all or parts of the code in your own programs
 - redistribute copies of the code
 - improve the code and release your improvements to the public
 However the use of the code is entirely your responsibility.
 */
package ch.aplu.ev3;

import com.jcraft.jsch.*;
import java.io.*;

/**
 * Class to download programs to the EV3 via SCP.
 * Based on the Java Secure Channel (Jsch) project from
 * http://www.jcraft.com/jsch/
 * You must include the jsch library in your class path.
 */
public class TargetCopy
{
  private boolean debug = false;

  /**
   * Class to provide user information for SSH login.
   */
  public static class MyUserInfo implements UserInfo
  {
    private String pwd = "";

    public MyUserInfo(String pwd)
    {
      this.pwd = pwd;
    }

    public String getPassword()
    {
      return pwd;
    }

    public boolean promptYesNo(String str)
    {
      return true;
    }

    public String getPassphrase()
    {
      return null;
    }

    public boolean promptPassphrase(String message)
    {
      return true;
    }

    public boolean promptPassword(String message)
    {
      return true;
    }

    public void showMessage(String message)
    {
    }
  }

  private final String VERSION = "1.21";
  private final int connectionTimeout = 10000; // 10 s
  private static String user;
  private static String pwd;
  private String ipAddress;
  private String sourcePath;
  private String targetPath;
  private String fileName;
  private String targetType;
  private final String errorMsg =
      "Check if the target brick is switched on,\n"
      + "and ready for connection over IP.";

  public TargetCopy(String ipAddress, String sourcePath, String fileName,
    String targetType)
  {
    System.out.println("TargetCopy V" + VERSION);
    if (targetType.equals("ev3"))
        System.out.println("Downloading '" + fileName + "' to EV3 (" + ipAddress + ")...");
    if (targetType.equals("raspi"))
        System.out.println("Downloading '" + fileName + "' to Raspberry Pi (" + ipAddress + ")...");
    
    this.ipAddress = ipAddress;
    this.sourcePath = sourcePath;
    this.targetType = targetType;
    this.fileName = fileName;

    if (this.targetType.equals("ev3"))
    {
      this.targetPath = "/home/python/scripts/" + fileName;
      user = "root";
      pwd = "";
    }
    else if (this.targetType.equals("raspi"))
    {
      this.targetPath = "/home/pi/scripts/MyApp.py";
      user = "pi";
      pwd = "raspberry";
    }
    else
    {
      System.out.println("Illegal target type");
      return;
    }

    if (debug)
    {
      System.out.println("Current values:");
      System.out.println("targetType: " + this.targetType);
      System.out.println("user: " + this.user);
      System.out.println("password: " + this.pwd);
      System.out.println("ipAddress: " + this.ipAddress);
      System.out.println("sourcePath: " + this.sourcePath);
      System.out.println("targetPath: " + this.targetPath);
      System.out.println("fileName: " + this.fileName);
    }
    /*
    System.out.println("Checking connection to " + ipAddress + "...");
    boolean rc = checkConnection();
    if (rc)
      System.out.println("Connection OK.");
    else
    {  
      System.out.println("Connection failed.\n" + errorMsg);
      isError = true;
    }
    */  
  }

  public boolean execute(String execScript)
  {
    if (targetType.equals("raspi"))
      return sendCommand("startApp");
    if (targetType.equals("ev3"))
      return sendCommand("pyrun /home/python/scripts/" + execScript);
    return false;
  }

  public boolean checkConnection()
  {
    JSch jsch = new JSch();
    String host = ipAddress;
    Session session = null;
    try
    {
      session = jsch.getSession(user, host, 22);
      UserInfo ui = new MyUserInfo(pwd);
      session.setUserInfo(ui);
      session.connect(connectionTimeout);
    }
    catch (JSchException ex)
    {
      System.out.println(ex.getMessage());
      return false;
    }
    session.disconnect();
    return true;
  }

  public boolean sendCommand(String cmd)
  {
    try
    {
      JSch jsch = new JSch();
      String host = ipAddress;
      Session session = jsch.getSession(user, host, 22);
      UserInfo ui = new MyUserInfo(pwd);
      session.setUserInfo(ui);
      session.connect();

      ChannelExec channel = (ChannelExec)session.openChannel("exec");
      BufferedReader in = new BufferedReader(new InputStreamReader(channel.getInputStream()));

      channel.setCommand(cmd);
      channel.connect();

      StringBuilder output = new StringBuilder();

      String s = null;
      while ((s = in.readLine()) != null)
      {
        output.append(s + "\n");
      }

      System.out.println(output);
      channel.disconnect();
      session.disconnect();
    }
    catch (Exception ex)
    {
      ex.printStackTrace();
      return false;
    }
    return true;
  }

  public boolean copy()
  {
    if (new File(sourcePath).length() == 0)
    {
      System.out.println("'" + fileName + "' empty. Nothing to download.");
      return false;
    }
    FileInputStream fis = null;
    try
    {
      String host = ipAddress;
      JSch jsch = new JSch();
      Session session = jsch.getSession(user, host, 22);

      UserInfo ui = new MyUserInfo(pwd);
      session.setUserInfo(ui);
      session.connect(connectionTimeout);

      boolean ptimestamp = true;

// exec 'scp -t remoteFile' remotely
      String command = "scp " + (ptimestamp ? "-p" : "") + " -t " + targetPath;
      Channel channel = session.openChannel("exec");
      ((ChannelExec)channel).setCommand(command);

// get I/O streams for remote scp
      OutputStream out = channel.getOutputStream();
      InputStream in = channel.getInputStream();

      channel.connect();

      if (checkAck(in) != 0)
      {
        displayError(1);
        return false;
      }

      File _lfile = new File(sourcePath);

      if (ptimestamp)
      {
        command = "T " + (_lfile.lastModified() / 1000) + " 0";
// The access time should be sent here,
// but it is not accessible with JavaAPI ;-<
        command += (" " + (_lfile.lastModified() / 1000) + " 0\n");
        out.write(command.getBytes());
        out.flush();
        if (checkAck(in) != 0)
        {
          displayError(2);
          return false;
        }
      }

// send "C0644 filesize filename", where filename should not include '/'
      long filesize = _lfile.length();
      command = "C0644 " + filesize + " ";
      command += fileName;
      command += "\n";
      out.write(command.getBytes());
      out.flush();
      if (checkAck(in) != 0)
      {
        displayError(3);
        return false;
      }

// send content of localFile
      fis = new FileInputStream(sourcePath);
      byte[] buf = new byte[1024];
      int totalLen = 0;
      int nbPoints = 0;
      while (true)
      {
        int len = fis.read(buf, 0, buf.length);
        if (len <= 0)
          break;
        out.write(buf, 0, len);
        totalLen += len;
        if (totalLen % 1000 == 0)
        {
          System.out.print(".");
          nbPoints++;
          if (nbPoints % 50 == 0)
            System.out.println();
        }
      }
      fis.close();
      fis = null;
// send '\0'
      buf[0] = 0;
      out.write(buf, 0, 1);
      out.flush();
      if (checkAck(in) != 0)
      {
        displayError(4);
        return false;
      }
      out.close();

      channel.disconnect();
      session.disconnect();
      System.out.println(" OK (" + totalLen + " bytes transferred).");
      return true;
    }
    catch (Exception e)
    {
      try
      {
        if (fis != null)
          fis.close();
      }
      catch (Exception ee)
      {
      }
      displayError(5);
      return false;
    }
  }

  private void displayError(int errNb)
  {
    System.out.println("\nDownload to " + ipAddress + " failed.\n"
      + errorMsg);
    if (debug)
      System.out.println("errNb: " + errNb);
  }

  private int checkAck(InputStream in) throws IOException
  {
    int b = in.read();
// b may be 0 for success,
// 1 for error,
// 2 for fatal error,
// -1
    if (b == 0)
      return b;
    if (b == -1)
      return b;

    if (b == 1 || b == 2)
    {
      StringBuffer sb = new StringBuffer();
      int c;
      do
      {
        c = in.read();
        sb.append((char)c);
      }
      while (c != '\n');
      if (b == 1)
      { // error
//        System.out.print(sb.toString());
      }
      if (b == 2)
      { // fatal error
//        System.out.print(sb.toString());
      }
    }
    return b;
  }
}
